/*
 * Linux EBSA285 BIOS IDE driver
 *
 * Bits taken from Linux kernel (C) respective authors.
 *
 * This is effectively a miniture IDE driver which is capable
 * of setting up the IDE interface for UDMA modes 0-2, DMA
 * multiword modes 0-2 or PIO modes 1 through 4.
 *
 * It also is capable of reading sectors off the IDE drive into
 * memory.
 *
 * IDE detection error codes:
 *   0 - success
 *   1 - unable to send IDENTIFY command
 *   2 - bad status from IDENTIFY command
 *   3 - unable to select drive (no interface?)
 */
#define NEW_PCI
#include <bios/boot.h>
#include <bios/time.h>
#include <bios/timer.h>
#include <bios/config.h>
#include <bios/bootdev.h>
#include <bios/malloc.h>
#include <bios/string.h>
#include <bios/pci.h>
#include <bios/stdio.h>

#include "ide.h"

#define NR_HW		8

/*
 * Wait 45 seconds after powerup for the hard drive to become
 * ready.
 */
#define MAX_BUSY_WAIT	45 * 100

static struct ide_hw	*ide_hw_list[NR_HW];

/*
 * Send a command to the drive and wait for BUSY status
 */
static inline int ide_send_command(struct ide_drive *drive, u8 cmd, int timeout)
{
	struct ide_hw *hw = drive->hw;

	ide_outb(hw, cmd, IDE_COMMAND);

	timeout += centisecs;

	do {
		if (timeout < centisecs)
			return 1;
		wait_cs(5);
	} while (ide_inb(hw, drive->statusreg) & BUSY_STAT);

	wait_cs(5);

	return 0;
}

/*
 * check whether status is ok
 */
static inline int ide_status_ok(struct ide_drive *drive, int ok, int bad)
{
	u8 s;

	s = ide_inb(drive->hw, IDE_STATUS);

	return (s & (ok | bad)) == ok;
}

/*
 * select a drive
 */
static inline int ide_select_drive(struct ide_drive *drive)
{
	ide_outb(drive->hw, drive->select, IDE_SELECT);

	wait_cs(5);

	return ide_inb(drive->hw, IDE_SELECT) == drive->select;
}

/*
 * Turn off the DMA capability status bit for this drive.
 * We also turn of the DMA enable bit.
 */
static void ide_dma_disable(struct ide_drive *drive)
{
	struct ide_hw *hw = drive->hw;
	int c;

	if (hw->dma_base) {
		c = ide_in_dma(hw, 2);
		if (drive->slave)
			c &= ~0x40;
		else
			c &= ~0x20;
		ide_out_dma(hw, c, 2);

		/*
		 * Ensure that the BM transfer bit is off.
		 */
		ide_out_dma(hw, ide_in_dma(hw, 0) & ~1, 0);
	}
}

/*
 * Turn on the DMA capability status bit for this drive.
 */
static void ide_dma_enable(struct ide_drive *drive)
{
	struct ide_hw *hw = drive->hw;
	int c;

	if (hw->dma_base) {
		c = ide_in_dma(hw, 2);
		if (drive->slave)
			c |= 0x40;
		else
			c |= 0x20;
		ide_out_dma(hw, c, 2);
	}
}

/*
 * read some data from the drive
 */
static void ide_input_data(struct ide_drive *drive, u8 *data, int len)
{
	unsigned short *s = (unsigned short *)data;

	while (len > 0) {
		len -= 2;
		*s++ = ide_inw(drive->hw, IDE_DATA);
	}
}

/*
 * convert a drive string to our endian-ness
 */
static void ide_fixstring(u8 *s, int size, int bswap)
{
	u8 *p = s, *end = &s[size & ~1];

	if (bswap) {
		for (p = end; p != s;) {
			unsigned short *pp = (unsigned short *)(p -= 2);
			*pp = *pp << 8 | *pp >> 8;
		}
	}

	while (s != end && *s == ' ')
		++s;

	while (s != end && *s) {
		if (*s++ != ' ' || (s != end && *s && *s != ' '))
			*p++ = *(s - 1);
	}

	/* wipe out trailing garbage */
	while (p != end)
		*p++ = '\0';
}

static int idx(unsigned int val, const unsigned int *vals)
{
	int i;

	for (i = 0; vals[i]; i++)
		if (val >= vals[i])
			break;

	return i;
}

static void get_speed_params(struct ide_drive *drive)
{
	struct hd_driveid *id = &drive->id;
	int eide_pio_cyc_time;

	if (drive->iordy)
		eide_pio_cyc_time = id->eide_pio_iordy;
	else
		eide_pio_cyc_time = id->eide_pio;

	drive->dma_speed = DMA_NONE;
	drive->pio_speed = 0;
	drive->dma = 0;

	/*
	 * Try UDMA modes 0 through 2
	 */
	if (ID_UDMA_VALID(id)) {
		if (id->dma_ultra & 4) {	/* UDMA mode 2 */
			drive->dma_speed = DMA_UDMA2;
			drive->dma = 2;
		} else
		if (id->dma_ultra & 2) {	/* UDMA mode 1 */
			drive->dma_speed = DMA_UDMA1;
			drive->dma = 2;
		} else
		if (id->dma_ultra & 1) {	/* UDMA mode 0 */
			drive->dma_speed = DMA_UDMA0;
			drive->dma = 2;
		}
	}

	/*
	 * Try EIDE DMA/timed PIO
	 */
	if (ID_PIODMA_VALID(id)) {
		/*
		 * EIDE DMA
		 */
		if (drive->dma == 0 && id->eide_dma_time != 0 && id->eide_dma_time <= 180) {
			static const unsigned int eide_times[] = { 150, 120, 90, 60, 0 };
			drive->dma = 1;

			drive->dma_speed = DMA_DMA0 + idx(id->eide_dma_time, eide_times);
		}
		/*
		 * EIDE PIO
		 */
		if (eide_pio_cyc_time != 0) {
			static const unsigned int eide_pio_times[] =
				{ 570, 540, 510, 480, 450, 420, 390, 360, 330,
				  300, 270, 240, 210, 180, 150, 120, 90, 60, 0 };

			drive->pio_speed = idx(eide_pio_cyc_time, eide_pio_times);
			return;
		}
	}

	switch (id->eide_pio_modes >> 8) {
	default:/* mode 4 - 210ns */
		drive->pio_speed = 13;
		break;

	case 1:	/* mode 3 - 390ns */
		drive->pio_speed = 7;
		break;

	case 0:
		drive->pio_speed = 0;
		break;
	}
}

static int set_drive_features(struct ide_drive *drive)
{
	int mode = 0;
	int failed = 0;

	switch (drive->dma) {
	case 2:	/* UDMA */
		mode = 0x40 | (drive->dma_speed - DMA_UDMA0);
		break;

	case 1:	/* DMA */
		if (drive->dma_speed == DMA_DMA1)
			mode = 0x21;
		else if (drive->dma_speed > DMA_DMA1)
			mode = 0x22;
		break;

	case 0: /* PIO */
		switch (drive->pio_speed) {
		case 16 ... 18:
			mode = 0x0c;	/* mode4 */
			break;

		case 14 ... 15:
			mode = 0x0b;	/* mode3 */
			break;

		default:
			break;
		}
		break;
	}

	if (mode) {
		if (!ide_select_drive(drive))
			failed = 1;

		ide_outb(drive->hw, mode, IDE_NSECTOR);
		ide_outb(drive->hw, 0x03, IDE_FEATURE);

		wait_cs(1);

		if (!failed && ide_send_command(drive, WIN_SETFEATURES, 2 * 100))
			failed = 1;

		if (!failed && !ide_status_ok(drive, 0, ERR_STAT|DRQ_STAT|BUSY_STAT))
			failed = 1;
	}

	return failed;
}

/*
 * Promise UDMA33 (PDC20246/7) interface initialisation.
 */
static void pdc20246_dma_init(struct ide_drive *drive, struct hd_driveid *id)
{
	static const struct { u8 c:4, b:3; } dma_regs[] = {
		{ 15, 7 },	/* NONE			*/
		{  4, 4 },	/* DMA0		180ns	*/
		{  4, 3 },	/* DMA1		150ns	*/
		{  3, 3 },	/* UDMA0 / DMA2	120ns	*/
		{  3, 3 },	/* UDMA0 / DMA2	120ns	*/
		{  2, 2 },	/* UDMA1	 90ns	*/
		{  1, 1 }	/* UDMA2	 60ns	*/
	};

	static const struct { u8 a:4, b:5; } pio_regs[] = {
		{ 13, 19 },	/* PIO0		600ns	*/
		{ 12, 18 },	/* PIO0		570ns	*/
		{ 11, 17 },	/* PIO0		540ns	*/
		{ 10, 16 },	/* PIO0		510ns	*/
		{  9, 15 },	/* PIO0		480ns	*/
		{  8, 14 },	/* PIO0		450ns	*/
		{  7, 13 },	/* PIO0		420ns	*/
		{  7, 12 },	/* PIO1		390ns	*/
		{  6, 11 },	/* PIO1		360ns	*/
		{  5, 10 },	/* PIO1		330ns	*/
		{  4,  9 },	/* PIO1		300ns	*/
		{  3,  8 },	/* PIO1		270ns	*/
		{  3,  7 },	/* PIO2		240ns	*/
		{  2,  6 },	/* PIO2		210ns	*/
		{  2,  5 },	/* PIO3		180ns	*/
		{  1,  4 },	/* PIO3		150ns	*/
		{  0,  3 },	/* PIO4		120ns	*/
		{  0,  2 },	/* PIO4		 90ns	*/
		{  0,  1 }	/* PIO4		 60ns	*/
	};
	struct ide_hw *hw = drive->hw;
	unsigned int base;
	u8 a, b, c;

	base = hw->ifnr ? 0x68 : 0x60;

	c = pci_read_config_byte(hw->dev, base + 2);
	if (c & 0x40)
		drive->iordy = 1;

	/*
	 * Calculate the UDMA/DMA/PIO capability for this drive
	 *  (results in drive->pio_speed, drive->dma_speed)
	 */
	get_speed_params(drive);

	if (!set_drive_features(drive)) {
		unsigned int drv_off;

		drv_off = base + (drive->slave ? 4 : 0);

		a = 0xd0 | pio_regs[drive->pio_speed].a;
		b = pio_regs[drive->pio_speed].b;
		c = 0;

		if (drive->iordy)
			a |= 0x20;

		if (drive->dma && drive->dma_speed <= DMA_UDMA2) {
			b |= dma_regs[drive->dma_speed].b << 5;
			c |= dma_regs[drive->dma_speed].c;
		} else {
			b |= 0xe0;
			c |= 0x0f;
		}

		pci_write_config_byte(hw->dev, drv_off + 0, a);
		pci_write_config_byte(hw->dev, drv_off + 1, b);
		pci_write_config_byte(hw->dev, drv_off + 2, c);

		if (drive->dma)
			ide_dma_enable(drive);

		switch (drive->dma) {
		case 2:	printf(", UDMA mode %d", drive->dma_speed - DMA_UDMA0); break;
		case 1: printf(", DMA mode %d", drive->dma_speed); break;
		case 0: printf(", PIO mode %d", drive->pio_speed >= 16 ? 4 :
						drive->pio_speed >= 14 ? 3 : 2); break;
		}
	}
}

#define pdc20246_init NULL

static int pdc20246_check(struct pci_dev *dev)
{
	unsigned int dma_base = dev->bar[4] & ~3;

	if (dma_base)
		pci_io_write_byte(0x01, dma_base + 0x1f);

	return 1;
}

static int pdc20262_check(struct pci_dev *dev)
{
	unsigned int dma_base = dev->bar[4] & ~3;

	if (dma_base)
		pci_io_write_byte(0x01, dma_base + 0x1f);

	return 1;
}

#define pdc20262_init NULL
#define pdc20262_dma_init NULL

/*
 * CY82C693 checking.  This has the primary IDE BARs on function 1, and
 * the secondary IDE BARs on function 2.  We therefore only recognise
 * function 1, and use dev->next to get at function 2.
 */
static int cy82c693_check(struct pci_dev *dev)
{
	return 0 && FUNC(dev->dev) == 1;
}

#define cy82c693_init NULL
#define cy82c693_dma_init NULL

/*
 * List of cards we know about...
 */
#define ENT(id,name,split,pre) \
  { id, name, split, pre##_check, pre##_init, pre##_dma_init }
static const struct dma_hw dma_hw[] = {
  ENT(0x105a4d33, "PDC20246", 0, pdc20246),
  ENT(0x105a4d38, "PDC20262", 0, pdc20262),
  ENT(0x1080c693, "CY82C693", 1, cy82c693),
  { 0, "unknown",  }
};

static int ide_lba_ok(struct hd_driveid *id)
{
	u32 lba_sects;
	u32 chs_sects;
	u32 chs_10_pc;

	/*
	 * ATA-6:
	 * If we have CHS of:
	 *  16383/16/63, 16383/15/63, 4092/16/63 or 4092/15/63
	 * and the lba capacity > 16383 * 63 * heads, then
	 * LBA mode must be supported.
	 */
	if (id->def_log_sects == 63 &&
            (id->def_log_cyls == 16383 ||
	     (id->def_log_cyls == 4092 && id->cur_log_cyls == 16383)) &&
	    (id->def_log_heads == 15 || id->def_log_heads == 16) &&
	    id->lba_capacity >= 16383*63*id->def_log_heads)
		return 1;

	/*
	 * ATA-2 check for LBA mode.  (note that this bit is
	 * "reserved" in ATA-6)
	 */
	if ((id->capability & 2) == 0)
		return 0;

	lba_sects = id->lba_capacity;
	chs_sects = id->def_log_cyls * id->def_log_heads * id->def_log_sects;
	chs_10_pc = chs_sects / 10;

	if ((lba_sects - chs_sects) < chs_10_pc)
		return 1;

	lba_sects = (lba_sects << 16) | (lba_sects >> 16);
	if ((lba_sects - chs_sects) < chs_10_pc) {
		id->lba_capacity = lba_sects;
		return 1;
	}

	return 0;
}

/*
 * Parse the IDENTIFY information
 */
static int ide_parse_identify(struct ide_drive *drive, struct hd_driveid *id)
{
	ide_fixstring(id->model, sizeof(id->model), 1);
	id->model[sizeof(id->model)-1] = '\0';

	if (id->def_log_sects && id->def_log_heads && id->def_log_cyls) {
		/*
		 * The drive supports CHS mode - get the default
		 * logical translation
		 */
		drive->sect = id->def_log_sects;
		drive->head = id->def_log_heads;
		drive->cyl  = id->def_log_cyls;

		/*
		 * Now check to see if the current logical translation
		 * is valid.  If it is, we should use this instead.
		 */
		if (ID_CURLOGGEO_VALID(id) && id->cur_log_cyls &&
		    id->cur_log_heads && id->cur_log_heads <= 16 &&
		    id->cur_log_sects) {
		    	u32 capacity, check;

			drive->sect = id->cur_log_sects;
			drive->head = id->cur_log_heads;
			drive->cyl  = id->cur_log_cyls;

			/*
			 * fix cur capacity if its wrong. ATA-6 defines
			 * this to be the product of the current logical
			 * translation CHS.
			 */
			capacity = drive->cyl * drive->head * drive->sect;
			check = id->cur_capacity0 << 16 | id->cur_capacity1;
			if (check == capacity) {
				id->cur_capacity0 = capacity;
				id->cur_capacity1 = capacity >> 16;
			}
		}

		drive->lba = 0;
		drive->capacity = drive->cyl * drive->head * drive->sect;
	} else {
		/*
		 * Drive supports only LBA mode
		 */
		drive->capacity = id->lba_capacity;
		drive->sect = 63;
		drive->head = 16;
		drive->cyl  = id->lba_capacity / (16 * 63);
		drive->lba  = 1;
	}

	if (!drive->lba && ide_lba_ok(id)) {
		/*
		 * Ok, the drive supports CHS mode and
		 * apparantly supports LBA mode?
		 */
		if (id->lba_capacity >= drive->capacity) {
			drive->cyl = id->lba_capacity / (drive->head * drive->sect);
			drive->lba = 1;
			drive->capacity = id->lba_capacity;
		}
	}

	/*
	 * Now set or clear the LBA bit in the head/device register
	 */
	if (drive->lba)
		drive->select |= 0x40;
	else
		drive->select &= 0xbf;

	printf("%s, %dMB, %sCHS=%d/%d/%d",
		id->model,
		drive->capacity / 2048,
		drive->lba ? "LBA, " : "",
		drive->cyl, drive->head, drive->sect);

	/*
	 * Do any chipset-specific DMA-type setup
	 */
	if (drive->hw->dma->setup_dma)
		drive->hw->dma->setup_dma(drive, id);

	printf("\n");
	return 0;
}

/*
 * Figure out which status register to use.  We try to use
 * the ALTSTATUS register in preference to the normal
 * STATUS register at all times.
 */
static inline int ide_detect_status_reg(struct ide_hw *hw)
{
	u8 s, a;

	a = ide_inb(hw, IDE_ALTSTATUS) & 0xc9;
	s = ide_inb(hw, IDE_STATUS)    & 0xc9;

	return (a ^ s) ? IDE_STATUS : IDE_ALTSTATUS;
}

/*
 * Handle IDENTIFY_DEVICE protocol
 */
static int ide_identify_device(struct ide_drive *drive, u8 cmd, void *buf)
{
	int rc = 1;

	if (ide_send_command(drive, cmd, 31 * 100))
		goto out;

	if (ide_status_ok(drive, DRQ_STAT, BAD_R_STAT)) {
		ide_input_data(drive, buf, 512);
		rc = 0;
	} else
		rc = 2;

	ide_inb(drive->hw, IDE_STATUS);
out:
	return rc;
}

static int ide_drive_probe(struct ide_drive *drive)
{
	int rc, count;

	/*
	 * Now, select which drive we want to talk to.  At
	 * this point, we should be able to read back the
	 * device/head (SELECT) register.  This test will
	 * trap out the truely non-present drives.
	 */
	if (!ide_select_drive(drive)) {
		rc = 3;
		goto out;
	}

	/*
	 * Now, have another go at guessing which status
	 * register we should use.  This may be different
	 * from before.
	 */
	drive->statusreg = ide_detect_status_reg(drive->hw);

	/*
	 * Try to identify the device.
	 */
	count = 0;
	do {
		rc = ide_identify_device(drive, WIN_IDENTIFY, &drive->id);
		if (!rc) {
			if (ID_INCOMPLETE(&drive->id))
				printf("(i) ");
			ide_parse_identify(drive, &drive->id);
			break;
		}
	} while (++count < 2);

out:
	return rc;
}

/*
 * Read the status/altstatus register
 */
static inline int ide_hw_read_status_reg(struct ide_hw *hw)
{
	return ide_inb(hw, hw->statusreg) & 0xf9;
}

/*
 * Try to detect a drive.  Note that this routine should NOT be
 * called within 2ms of the negation of RESET to the drive.  Note
 * also that "centisecs" is the time past since we started running
 * the bios, so we can do an absolute timeout here.
 */
static int ide_hw_wait_ready(struct ide_hw *hw)
{
	int target = MAX_BUSY_WAIT, rc;

	/*
	 * First, select which status register to use
	 */
	hw->statusreg = ide_detect_status_reg(hw);

	/*
	 * Check the busy bit.  If it is set, we can't access any registers.
	 * We wait a maximum of 32 seconds for the status register to become
	 * not busy.  Note that the T13 ATAPI standard defines there to be
	 * a 10K pull-down resistor on this line.
	 */
	while (ide_hw_read_status_reg(hw) & BUSY_STAT) {
		if (centisecs > target) {
			rc = 2;
			goto out;
		}
		/*
		 * Wait a full second before trying again - we don't
		 * want to hammer the drive with status reads
		 */
		wait_cs(100);
	}

	/*
	 * Detect any error that the drive may be reporting.  Our parent will
	 * report the error code to the user.  Maybe we ought to tell the
	 * user in plain english which drive has failed?
	 */
	if (ide_hw_read_status_reg(hw) & ERR_STAT) {
		rc = 1;
		goto out;
	}
	rc = 0;
out:
	return rc;
}

/*
 * Initialise one interface
 */
static void ide_init_one_if(struct ide_hw *hw)
{
	int drvnr, rc, drvok = 3;

	printf(" ide%d: %s at 0x%04x,0x%04x",
		hw->ifnr, hw->dma->name, hw->regs[0], hw->regs[8]);
	if (hw->dma_base)
		printf(", BM-DMA at 0x%04x", hw->dma_base);
	if (hw->dma && hw->dma->init)
		hw->dma->init(hw);
	printf("\n");

	/*
	 * Wait for the interface to become not busy.  This can take
	 * up to 32 seconds after power on for the drive to finish
	 * its diagnostics, during which time the interface must
	 * otherwise remain idle.
	 */
	rc = ide_hw_wait_ready(hw);
	if (rc) {
		int error;
		switch (rc) {
		case 2:
			printf(" ide%d: timeout waiting for busy to "
				"deassert\n", hw->ifnr);
			drvok = 0;
			break;

		case 1:
			/*
			 * There are some drives that indicate an error
			 * because they have not detected a slave.  If
			 * this is the case, then we continue anyway.
			 * [ note that these drives typically don't
			 *   perform diagnostics on reset either ]
			 */
			error = ide_inb(hw, IDE_ERROR);
			if ((error & 0x7f) != 0x01) {
				printf(" ide%d: master drive diagnostic "
					"failure, code 0x%02x\n",
					hw->ifnr, error & 0x7f);
				drvok &= ~1;
			}
			if (error & 0x80) {
				/*
				 * Slave error - we must read its error
				 * register to get the full picture.
				 */
				ide_outb(hw, 0xb0, IDE_SELECT);
				error = ide_inb(hw, IDE_ERROR);
				ide_outb(hw, 0xa0, IDE_SELECT);
				printf(" ide%d: slave drive diagnostic "
					"failure, code 0x%02x\n",
					hw->ifnr, error);
				drvok &= ~2;
			}
		}
	}

	/*
	 * setup each drive on this interface, and
	 * then try to detect it.
	 */
	for (drvnr = 0; drvnr < 2; drvnr++) {
		struct ide_drive *drv;

		if (drvok & (1 << drvnr)) {
			drv = malloc(sizeof(struct ide_drive));
			memset(drv, 0, sizeof(struct ide_drive));

			hw->drive[drvnr] = drv;
			drv->hw     = hw;
			drv->slave  = drvnr;
			drv->name   = 'a' + hw->ifnr * 2 + drvnr;
			drv->select = 0xa0 | (drvnr << 4);

			ide_dma_disable(drv);

			printf("  hd%c: ", drv->name);
			rc = ide_drive_probe(drv);
			if (rc) {
				printf("not present (%d)\n", rc);
				hw->drive[drvnr] = NULL;
				free(drv);
			}
		}
	}
}

static void ide_init_hw(unsigned int reg_base, unsigned int ctrl_base,
			unsigned int dma_base, struct pci_dev *dev,
			const struct dma_hw *dma)
{
	static int ifnr;
	struct ide_hw *hw;
	int i;

	if (ifnr >= NR_HW || reg_base == 0)
		return;

	ide_hw_list[ifnr] = hw = malloc(sizeof(struct ide_hw));
	memset(hw, 0, sizeof(struct ide_hw));

	for (i = 0; i < 8; i++)
		hw->regs[i] = reg_base + i;

	hw->regs[8]  = ctrl_base;
	hw->dma_base = dma_base;
	hw->dev      = dev;
	hw->dma      = dma;
	hw->ifnr     = ifnr;

	ide_init_one_if(hw);

	ifnr += 1;
}

static void ide_init_one_hw(struct pci_dev *dev, const struct dma_hw *dma)
{
	int i;

	for (i = 0; i < 2; i++) {
		unsigned int reg_base, ctrl_base, dma_base;

		if (i == 0) {
			reg_base  = dev->bar[0] & ~3;
			ctrl_base = dev->bar[1] & ~3;
			dma_base  = dev->bar[4] & ~3;
		} else if (!dma->split_cfg) {
			reg_base  = dev->bar[2] & ~3;
			ctrl_base = dev->bar[3] & ~3;
			dma_base  = (dev->bar[4] & ~3) + 8;
		} else {
			reg_base  = dev->next->bar[0] & ~3;
			ctrl_base = dev->next->bar[1] & ~3;
			dma_base  = (dev->bar[4] & ~3) + 8;
		}
		ide_init_hw(reg_base, ctrl_base + 2, dma_base, dev, dma);
	}
}

/*
 * Probe for all IDE devices on this bus.
 */
static int ide_probe(void)
{
	struct pci_dev *dev = NULL;

	/*
	 * Search for known interfaces
	 */
	while ((dev = pci_dev_next(dev)) != NULL) {
		const struct dma_hw *dma;
		int i, ok = 0;

		/*
		 * If this is an IDE or RAID storage adapter,
		 * then we have found the correct entry.
		 */
		if (dev->class == PCI_CLASS_STORAGE_IDE ||
		    dev->class == PCI_CLASS_STORAGE_RAID)
			ok = 1;

		/*
		 * Now look it up in the table of known adapters - we
		 * may have to ignore this entry for some reason.
		 */
		for (i = 0, dma = dma_hw; dma->vendev; i++, dma++) {
			if ((dev->vendor << 16 | dev->device) == dma->vendev) {
				ok = dma->check ? dma->check(dev) : 1;
				break;
			}
		}

		if (ok)
			ide_init_one_hw(dev, dma);
	}

	return 0;
}

static int ide_start(void)
{
	int ifnr, drvnr, present = 0;

	for (ifnr = 0; ifnr < NR_HW; ifnr++) {
		struct ide_hw *hw = ide_hw_list[ifnr];

		/*
		 * Interface present?
		 */
		if (!hw)
			continue;

		/*
		 * Try each drive
		 */
		for (drvnr = 0; drvnr < 2; drvnr++) {
			if (hw->drive[drvnr])
				present = 1;
		}
	}

	return present ? 0 : -1;
}

#if 0
static void dump_sector(u8 *buffer)
{
	int i, j;

	for (i = 0; i < 512; i += 16) {
		debug_printf("%03X  ", i);

		for (j = i; j < i + 16; j++)
			debug_printf("%02X ", buffer[j]);

		debug_printf("  ");

		for (j = i; j < i + 16; j++)
			debug_printf("%c", (buffer[j] < 32 || buffer[j] > 127) ? '.' : buffer[j]);
		debug_printf("\n");
	}
}
#endif

static int __ide_read_sectors(struct ide_drive *drive, u32 sector, u8 *buf, int nsect)
{
	u8 sect, lcyl, hcyl, sel;
	unsigned int target;
	int status, rc = -1;

	if (sector > drive->capacity - nsect) {
		printf("__ide_read_sectors: bad sector 0x%x len 0x%x > capacity 0x%x\n",
			sector, nsect, drive->capacity);
		return -1;
	}

	if (!ide_status_ok(drive, 0, BUSY_STAT)) {
		printf("__ide_read_sectors: drive hd%c busy\n",
			drive->name);
		return -1;
	}

	if (!ide_select_drive(drive))
		return -1;

	wait_cs(1);

	if (!ide_status_ok(drive, READY_STAT, BUSY_STAT|DRQ_STAT)) {
		printf("hd%c: drive not ready for command\n", drive->name);
		return -1;
	}

	ide_outb(drive->hw, 8, IDE_CONTROL);
	ide_outb(drive->hw, nsect, IDE_NSECTOR);

	if (drive->lba) {
		sect = sector;
		lcyl = sector >> 8;
		hcyl = sector >> 16;
		sel  = (sector >> 24) & 15;
	} else {
		unsigned int cyl, track;

		sect  = sector % drive->sect + 1;
		track = sector / drive->sect;
		sel   = track % drive->head;
		cyl   = track / drive->head;
		lcyl  = cyl;
		hcyl  = cyl >> 8;
	}

	ide_outb(drive->hw, sect, IDE_SECTOR);
	ide_outb(drive->hw, lcyl, IDE_LCYL);
	ide_outb(drive->hw, hcyl, IDE_HCYL);
	ide_outb(drive->hw, sel | drive->select, IDE_SELECT);
	ide_outb(drive->hw, WIN_READ, IDE_COMMAND);

	wait_cs(1);

	rc = -2;
	target = centisecs + 2000;
	do {
		status = ide_inb(drive->hw, IDE_STATUS);

		if (status & BUSY_STAT)
			continue;

		if (status & DRQ_STAT) {
			target = centisecs + 100;
			ide_input_data(drive, buf, 512);
			buf += 512;
			nsect -= 1;
		}

		if (status & ERR_STAT)
			break;
	} while (target > centisecs && nsect);

	if (status & ERR_STAT) {
		printf("hd%c: error reading sectors: 0x%02x\n",
			drive->name, ide_inb(drive->hw, IDE_ERROR));
	} else if (!nsect)
		rc = 0;

	return rc;
}

static int ide_read_sectors(struct ide_drive *drive, u32 sector, void *buffer, int len)
{
	u8 *buf = (u8 *)buffer;
	int rc = -1;

	if (len & 511)
		return -1;

	len = len >> 9;

	while (len) {
		int nsect;

		nsect = len;
		if (nsect > 244)
			nsect = 244;

		rc = __ide_read_sectors(drive, sector, buf, nsect);

		if (rc)
			break;

		len    -= nsect;
		sector += nsect;
		buf    += (nsect << 9);
	}

	return rc;
}

static u32 part_getbootstart(struct ide_drive *drive)
{
	u8 sector[512], *p;
	int nr;

	if (ide_read_sectors(drive, 0, sector, 512))
		return 0;

	if ((sector[510] != 0x55 || sector[511] != 0xaa) &&
	    (sector[510] != 'R' || sector[511] != 'K'))
		return 0;

	for (p = sector + 0x1be, nr = 0; nr < 4; p += 16, nr++)
		if (p[0] == 0x80)
			return p[8] | p[9] << 8 | p[10] << 16 | p[11] << 24;

	return 0;
}

struct map {
	u32	magic;
	u32	block_sz;
	struct {
		u32	off;
		u32	len;
	} map[63];
};

static int read_mapped_file(struct ide_drive *drive, u32 part_start, struct map *map, u8 *ptr)
{
	int i;

	if (map->magic != 0xc53a4b2d)
		return -1;

	for (i = 0; i < 63 && map->map[i].off; i++) {
		u32 start;

		start = part_start + map->map[i].off * map->block_sz / 512;

		if (ide_read_sectors(drive, start, ptr, map->map[i].len * map->block_sz))
			return -1;

		ptr += map->map[i].len * map->block_sz;
	}

	return 0;
}

static void copy_map(struct map *map, u8 *from)
{
	u32 *p = (u32 *)from;
	int i = 0;

	map->magic = *p++;
	map->block_sz = *p++;

	while (p[0] && p[1]) {
		map->map[i].off = *p++;
		map->map[i].len = *p++;
		i += 1;
	}
}

extern char *cmdline;

static int ide_load(void)
{
	struct ide_drive *drive = NULL;
	struct map map;
	u8 *boot_map_file = (u8 *)0x4000;
	u32 part_start = 0;
	u32 *p;
	int nr, i, bmf_size;
	extern int img_nr;

	/* If J17 P13-14 is made, then don't boot off the HD
	 */
	if (!((*(u8 *)0x40012000) & 0x10))
		return -1;

	for (i = 0; i < NR_HW; i++) {
		struct ide_hw *hw = ide_hw_list[i];
		int drvnr;

		if (!hw)
			continue;

		for (drvnr = 0; drvnr < 2; drvnr ++) {
			drive = hw->drive[drvnr];

			if (drive) {
				debug_printf("checking hd%c: ", drive->name);
				part_start = part_getbootstart(drive);
				debug_printf("boot start %08x\n", part_start);
			}

			if (part_start)
				break;
		}
		if (part_start)
			break;
	}

	if (!drive || part_start == 0)
		return -1;

	debug_printf("Using hd%c...\n", drive->name);

	if (ide_read_sectors(drive, part_start, &map, sizeof(map)))
		return -1;

	debug_printf("read boot sectors: magic = %08x\n", map.magic);

	if (read_mapped_file(drive, part_start, &map, boot_map_file))
		return -1;

	debug_printf("read map sectors: magic = %08x\n", *(u32 *)boot_map_file);

	if (*(u32 *)boot_map_file != 0xc35ae183)
		return -1;

	bmf_size = boot_map_file[4];
	nr = boot_map_file[8];

	printf("Available configurations:\n");

	for (i = 0; i < nr; i++) {
		p = (u32 *)(boot_map_file + bmf_size * (i + 1));

		if (p[0] != 1)
			continue;

		printf("  %d %-10s\n", i + 1, boot_map_file + p[1]);
	}

	if (img_nr >= nr || img_nr < 0)
		img_nr = 0;

	p = (u32 *)(boot_map_file + bmf_size * (img_nr + 1));

	printf("Loading %s...", boot_map_file + p[1]);

	if (p[2])
		cmdline = boot_map_file + p[2];

	copy_map(&map, boot_map_file + p[3]);

	if (read_mapped_file(drive, part_start, &map, (u8 *)load_addr))
		return -1;

	if (root_dev == 0)
		root_dev = p[5];
	root_flags = p[7];

	return 0;
}

static int ide_stop(void)
{
	return 0;
}

struct bootdev boot_ide = {
	"ide",

	ide_probe,
	ide_start,
	ide_load,
	ide_stop
};
